structure Server : SERVER =
struct
        open Common
        open Types
        exception KillRobot

        (* Many of these are sent to a client after a turn. *)
        datatype event = MoveEvent of Common.dir | DropEvent of Common.packid
          | PickEvent of Common.packid | AppearEvent of Common.loc


	local 
	    fun ++ (r as (ref z)) = (r := (z + 1); z + 1)
	    val r = ref 0
	    val bad = ref (Word32.fromLargeInt (Time.toSeconds (Time.now())))
		
	    fun random () =
		let in
		    bad := Word32.*(!bad, 0w177801);
		    bad := Word32.xorb(0wxF3014823, !bad);
		    bad := Word32.+(!bad, 0w9999111);
		    bad := Word32.>>(!bad, 0w1);
		    !bad
		end
	in
	    fun random_bit () = (Word32.andb(random (), 0wx100) > 0w0)
	    fun newroboid () = ++ r
	end

        (* These are the actual messages sent; they come with the robot ID that
           the event pertains to. *)
        type response = (Common.robotid * event list) list

        val its = Int.toString
        fun sti x =
            (case Int.fromString x of
                 SOME y => y
               | NONE => raise KillRobot)
                                 
        fun stringify_dim (w : world) = (its(#width w-2)) ^ " " ^ 
                                        (its(#height w-2))

        fun init_response (w : world) = map (fn (r : robot) => 
                                             (#id r, [AppearEvent(#pos r)])) (#robots w)

        fun stringify_event (MoveEvent NN) = "N"
          | stringify_event (MoveEvent EE) = "E"
          | stringify_event (MoveEvent SS) = "S"
          | stringify_event (MoveEvent WW) = "W"
          | stringify_event (DropEvent pl) = "D " ^ its pl
          | stringify_event (PickEvent pl) = "P " ^ its pl
          | stringify_event (AppearEvent(x, y)) = "X " ^ its x ^ " Y " ^ its y

        fun stringify_response r = (StringUtil.delimit " " 
                (map (fn (rid, el) => "#" ^ its rid ^ " " ^ StringUtil.delimit " " (map stringify_event el)) r))

        val stringify_spot =
            fn WALL => "#"
             | EMPTY => "."
             | WATER => "~"
             | BASE => "@"

        fun stringify_board ({board, width, height, ...}:world) =
            Array.foldli (fn (i, spot, buf) =>
                          let
                              val col = i mod width
                              val row = i div width
                          in
                              if row = 0 orelse row = height-1 orelse col = 0 orelse col = width-1 then
                                  buf
                              else if col = width-2 andalso row < height-2 then
                                  buf ^ stringify_spot spot ^ "\n"
                              else
                                  buf ^ stringify_spot spot
                          end) "" (board, 0, NONE)
                              

	fun robo_init ({money, id, ...} : robot, {cap, ...} : world) =
	    StringUtil.delimit " " (map its [id, cap, money])

        fun send_init (r : robot, w : world) = 
	    let 
                val send = LSocket.send (#sock r)
	    in
                send (stringify_dim w);
                send (stringify_board w);
		send (robo_init (r,w));
                send (stringify_response(init_response w))
	    end

        fun get_cmd ({sock,...}:robot) =
            let
                val line = LSocket.recv sock
                val tokens = String.tokens Char.isSpace line
            in
                case tokens of
                    [n, "Move", "N"] => (sti n, Move NN)
                  | [n, "Move", "E"] => (sti n, Move EE)
                  | [n, "Move", "W"] => (sti n, Move WW)
                  | [n, "Move", "S"] => (sti n, Move SS)
                  | n::"Pick"::ps => (sti n, Pick (map sti ps))
                  | n::"Drop"::ps => (sti n, Drop (map sti ps))
                  | _ => raise KillRobot
            end

        fun find_packages ({robots,packages,...}:world) : (robot * package list) list =
            map (fn robot as {pos = (row:int, col:int), id, ...} =>
                 (robot, List.filter (fn (_, _, NONE, _) => false
                                   | (_, _, SOME (crow:int, ccol:int), _) => row = crow andalso col = ccol) packages)) robots

        fun stringify_package (id, weight, _, (row, col)) = its id ^ " " ^ its col ^ " " ^ its row ^ " " ^ its weight

        fun send_packages ({sock,...}:robot, ps) =
            LSocket.send sock (StringUtil.delimit " " (map stringify_package ps))

        val stringify_dir =
            fn NN => "N"
             | EE => "E"
             | SS => "S"
             | WW => "W"

        val stringify_event =
            fn MoveEvent dir => stringify_dir dir
             | DropEvent id => "D " ^ its id
             | PickEvent id => "P " ^ its id
             | AppearEvent (row, col) => "X " ^ its col ^ " Y " ^ its row

        fun send_responses ({sock,...}:robot, resps) =
            LSocket.send sock (StringUtil.delimit " " (map (fn (id, evs) =>
                                                             "#" ^ its id ^ " " ^ 
							     StringUtil.delimit " " (map stringify_event evs)) resps))

	fun kill_robot (s : string) ({id, sock, score, ...} : robot) =
	    let in
		LSocket.hangup sock;
		print ("Killing Robot " ^ its id ^ ": " ^ s ^ " (score: " ^ its score ^ ")\n")
	    end


        fun boardat ({board, width, height, ... }:world) row col = Array.sub(board, row * width + col)

        fun step cmds {board, robots, cap, height, width, packages} =
	    let
	       (* annihilate over-bidding robots. *)
		val newcmds = List.filter (fn (r as {money,...}, (bid, _)) =>
					   if Int.abs bid > money orelse bid = 0 orelse bid > 50 orelse bid < ~50 then
					       (kill_robot "illegal bid" r; false)
					   else true) cmds
		val world = { robots = map #1 newcmds, cap = cap, packages = packages, 
			      height = height, width = width, board = board }
		(* unstable! sort commands by bid *)
		fun compare ((_, (b1, _)), (_, (b2, _))) =
		     case Int.compare (b1, b2) of
			     EQUAL => if random_bit () then LESS else GREATER
	                   | order => order

		val cmds = ListUtil.sort compare newcmds

		fun get_pkg i nil = raise Fail "impossible"
		  | get_pkg i ((id, a, b, c)::rest) = if i = id then (a,b,c)
						      else get_pkg i rest
							  
		fun pts (r, c) = "(" ^ its r ^ ", " ^ its c ^ ")"

		fun robot_lookup i ((r as {id, ...}: robot) :: rest) = if i = id then r else robot_lookup id rest
		  | robot_lookup i nil = raise Fail "impossible"

		fun update_robots (nr as {id, ...}) ((or as {id=ido, ...} : robot) :: rest) =
		    if id = ido then nr :: rest
		    else or :: update_robots nr rest
		  | update_robots _ _ = raise Fail "impossible"

		(* we modify board in place, so just pass list of robots and packages *)
		fun execute (({id=rid,...} : robot, (bid, cmd))::restcmds) robots packages reboots resps =
		    (* FIXME filter out the reboots *)
		    let
			val r = robot_lookup rid robots
			val rp = #packages r
			val pos = #pos r
		    in (case cmd of
			Pick l =>
			    let
				val _ = print ("rob " ^ its (#id r) ^ " pick up at " ^ its (#1 pos) ^ "," ^ its (#2 pos) ^ "\n")
				val cweight = List.foldl (fn (pid, ww) =>
							  let val (w,_,_) = get_pkg pid packages
							  in ww + w
							  end) 0 rp
				val ok = List.filter (fn i => List.exists (fn (j,w,SOME loc,_) => 
									   let in
									       print ("pkg " ^ its j ^ " weighs " ^ its w ^
										      " at " ^ its (#1 loc) ^ "," ^ its (#2 loc) ^
										      "\n");
									       i = j andalso loc = pos
									       andalso w + cweight <= cap
									   end
									   | _ => false) packages) l
				val _ = app (fn id => print ("Robot " ^ its (#id r) ^ " pick up " ^ its id ^ "\n")) ok
				val held = ok @ rp
				val packages = List.map (fn (s as (id, w, _, d)) => if List.exists (fn j => j = id) ok 
											     then (id, w, NONE, d)
											 else s) packages

				val newrob = { packages = held,
					       score = #score r,
					       pos = #pos r,
					       money = #money r - Int.abs bid,
					       sock = #sock r,
					       id = #id r } : robot
			    in
				execute restcmds (update_robots newrob robots) packages reboots ((#id r, map PickEvent ok) :: resps)
			    end
		      | Drop l =>
			    let
				val (dropped, stay) = List.partition (fn i => List.exists (fn j => j = i) l) rp

				fun handledrop (pid::rest) =
				    let val (weight, _, dest) = get_pkg pid packages
					val (sc, rem, dp) = handledrop rest
					val _ = print (pts dest ^ " vs. " ^ pts pos ^ "\n")
				    in if dest = pos then (sc + weight, pid :: rem, dp)
				       else (sc, rem, pid :: dp)
				    end
				  | handledrop nil = (0, nil, nil)

				val (score, rem, landed) = handledrop dropped

				val packages = List.filter (fn (id,_,_,_) => List.all (fn j => j <> id) rem) packages

				fun replace_pgks (p as (id, w, NONE, dst)) =
				    if List.exists (fn j => j = id) landed then
					(id, w, SOME pos, dst)
				    else p
				  | replace_pgks p = p

				val packages = map replace_pgks packages

				val newrob = {packages=stay, score = #score r + score, pos = #pos r, 
					      money = #money r - Int.abs bid, sock = #sock r, id = #id r } : robot

			    in execute restcmds (update_robots newrob robots) 
				  packages reboots ((#id r, map DropEvent dropped) :: resps)
			    end
		      | Move d => 
			    let
				val (dr, dc) = case d of NN => (~1, 0) | SS => (1, 0) | WW => (0, ~1) | EE => (0, 1)
				(* find the robots along our ray, push them. Return a list of newly rebooted robots,
				   and a bool indicating if we can move forward, the new list of packages,
                                   and the whole list of robots, and responses *)

				fun getbotsat loc robs = List.filter (fn ({pos, ...} : robot) => loc = pos) robs

				fun findbots (r, c) =
				    (case boardat world r c of
					 WALL => (nil, false, packages, robots, nil)
				       | WATER => (nil, true, packages, robots, nil)
				       | _ => (case getbotsat (r, c) robots of
						     nil => (nil, true, packages, robots, nil)
						   | botshere =>
							 let
							     val (rbts, free, packs, robs, resps) = findbots (r + dr, c + dc)
							     fun dropmove nil = (nil, nil)
							       | dropmove (({packages=rp, pos=(pr,pc), id=this, ...} 
									    : robot)::rest) =
								 let
								     val newloc = (pr + dr, pc + dc)

								     fun dodrop nil = (nil, nil, packs)
								       | dodrop (pid::rest) =
									 if random_bit () then
									     ([(this, DropEvent pid)],
									      rest,
									      map (fn (p as (id, w, NONE, d)) =>
										   if id = pid then (id, w, SOME (r, c), d)
										   else p
									           | p => p) packs)
									 else dodrop (rest @ [pid])
								     val (newresps,
									  robotpkgs,
									  worldpacks) = dodrop rp
								 in
								     raise Match
								 end
							 in
							     raise Match
							 end))
			    in
				raise Match
			    end)
		    end
		  | execute nil robdone packages _ resps = (robdone, packages, resps)

		val (robdone, packages, resps) = execute cmds (map #1 cmds) packages nil nil
	    in
		({robots = robdone, packages = packages, board = board, width = width, height = height, cap = cap}, resps)
	    end

        (* return a valid position for a new robot 
           XXX - make better *)
        fun place_robot (existing : robot list) (world as { board, width, height, ... } : world) = 
            case (Util.ford 1 (width-2) (~1, ~1)
                  (fn (col, pr) =>
                   (Util.ford 1 (height-2) pr
                    (fn (row, (i,j)) =>
                     case boardat world row col of
                         EMPTY => if List.exists (fn {pos = (r, c), ...} => r = row andalso c = col) existing then
                                  (i, j) else (row, col)
                       | _ => (i, j))))) of
            (~1,_) => (print "No more room !\n"; raise KillRobot)
          | loc => loc

	fun wait_player svr =
	    let val peer = LSocket.accept svr
		val _ = print "\nGot connection...\n\n"
		val msg = LSocket.recv peer
	    in
		print ("Got [" ^ msg ^ "]\n");
		msg = "Player" orelse (LSocket.hangup peer; raise KillRobot);
		peer
	    end
            
	fun read_command (world as {packages,...} : world) (robot as {pos,sock,...} : robot) =
	    let 
		val lpackages = List.filter (fn (_, _, SOME loc, _) => loc = pos
					      | _ => false) packages
	    in
		LSocket.send sock 
		     (StringUtil.delimit " " (map (fn (id, w, _, (row,col)) => 
						   (its id ^ " " ^ its col ^ " " ^ its row ^ " " ^ its w)) lpackages));
		(SOME (robot, get_cmd robot)) handle KillRobot =>
		    let in
			kill_robot "malformed command" robot;
			NONE
		    end
	    end

	fun game_over {robots=nil, ...} = true
	  | game_over {packages=nil, ...} = true
	  | game_over (_ : world) = false

	fun loop (world as {robots, cap, width, height, packages, board} : world) =
	    let
		(* read commands *)
		val cmds = List.mapPartial (read_command world) robots
		(* update world; some robots are dead *)
		val world = {robots=map #1 cmds, cap=cap, width=width, 
			     height=height, packages=packages, board=board}
		(* step world *)
		val (new_world, msgs) = step cmds world
		(* send responses -- only to remaining robots *)
		val _ = app (fn r => send_responses (r, msgs)) (#robots new_world)
		(* check termination condition *)
	    in
		if game_over new_world then
		    let in
			print "The game is over!!\n";
			app (fn {sock, ...} => LSocket.hangup sock) (#robots new_world)
		    end
		else loop new_world
	    end

	fun really_go (world as {robots, ...} : world) =
	    let
		(* send board to everyone *)
		val _ = app (fn r => send_init (r, world)) robots
	    in
		loop world
	    end
	    
        fun go map port robots cap cash =
            let
		val _ = print "Welcome to the ARENA ETERNAL \n"
                val server = LSocket.listen port
                val Parser.FILE {world = w as {board, packages, width, height, ...},
                                 spawns} = Parser.parse map
            in
              really_go
              { board=board, packages=packages, width=width, height=height, cap=cap,
               robots=Util.ford 1 robots nil (fn (i, l) => {pos=place_robot l w,
                                                            id=newroboid (),
                                                            packages=nil,
                                                            score=0,
                                                            money=cash,
                                                            sock=wait_player server}::l)
               }
            end
          handle e => (TextIO.output (TextIO.stdErr, String.concat ["go failed: ",
                                                                    exnMessage e,
                                                                    "\n"]))

end
