structure Comm :> COMM =
  struct
    open World

    val pause = Params.flag false (SOME ("-pause", 
					"Pause after every move.")) "pause"

    val debug = Params.flag false (SOME ("-debug", 
					 "Turn on debugging messages.")) "debug"

    val showboard = Params.flag false (SOME ("-showboard", 
					 "Show the board.")) "showboard"

    type bid = int

    fun pr s = TextIO.output (TextIO.stdErr, String.concat s)

    fun recv s = let val x = LSocket.recv s in if !debug then (pr ["LSocket: recv'd \"", x, "\"\n"]) else (); x end
(*     val recv = Util.time "recv" recv *)

    val chars = Vector.fromList (List.map Char.toString 
				 (String.explode "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz!$%&'()*+'-./"))
    val charslen = Vector.length chars

    fun stringify_spot (World.SPOT {space = w, ...}) =
      case w of 
        World.Wall => "\027[0;41m\027[1;31m#\027[0;40m\027[0;37m"
      | World.Dot false => "\027[1;30m.\027[0;37m"
(*       | World.Dot false => "." *)
      | World.Water => "\027[0;34m~\027[0;37m"
      | World.Dot true => "\027[1;36m@\027[0;37m"
      
    val itos = Int.toString
    fun dtos (a,b) = itos a ^ "," ^ itos b
    fun pltos (World.Ground d) = "@" ^ dtos d
      | pltos (World.Robot r) = "[rob " ^ itos r ^ "]"

    fun ptos (World.PACKAGE{id,loc,weight,dest}) = itos id

    fun ktos f (World.Known n) = f n
      | ktos _ _ = "Unknown"

    fun pr_board (World.BOARD {layout = (board, width, height), 
                                      robots, packages, ...}) =
      let
        val roblocs = List.map (fn (ri, World.ROBOT {loc, ...}) => (ri, loc)) (World.Map.listItemsi robots)
      in
        (*pr ["\027[H\027[2J"];*)
        Array.foldli (fn (i, _, ()) =>
                      let
                        val col = i mod width
                        val row = (height - 1) - i div width
                        val spot = Array.sub (board, col + row * width)
                      in
                        if row = 0 orelse row = height-1 orelse col = 0 orelse col = width-1 then
                           ()
                         else
                           (case List.find (fn (ri, (r, c)) => r = row andalso c = col) roblocs of
                               SOME (ri, (r, c)) => pr["\027[1;33m", Vector.sub (chars, ri mod charslen), "\027[0;37m",
						       if col = width-2 then "\n" else ""]
                            | NONE => pr [stringify_spot spot, if col = width-2 then "\n" else ""])
                      end) () (board, 0, NONE);
	if !debug then
       (pr ["Packages of " ^ itos (World.Map.numItems robots) ^ " robots follow:\n"];
        pr [World.Map.foldri (fn (id, World.ROBOT {packages = robpacks, ...}, str) =>
			      str ^
                          "Robot " ^ itos id ^ "(" ^
			  Vector.sub(chars, id mod charslen) ^
			  ")'s packages:\n     " ^ StringUtil.delimit "\n     "
                          (map (fn p => (case World.Map.find (packages, p) of
                                           SOME (World.PACKAGE{id, loc, weight, dest}) =>
                                             itos id ^ ": " ^ ktos dtos dest ^ " " ^ ktos itos weight ^ "lb. " ^ ktos pltos loc
                                         | NONE => "")) robpacks) ^ "\n") "" robots]) else ()
      end

    fun dirname N = "N"
      | dirname S = "S"
      | dirname E = "E"
      | dirname W = "W"

    fun seract (Move dir) = "Move " ^ dirname dir
      | seract (Pick pl) = 
        "Pick " ^ (StringUtil.delimit " " o List.map (fn World.PACKAGE {id, ...} => Int.toString id)) pl
      | seract (Drop pl) =
        "Drop " ^ (StringUtil.delimit " " o List.map (fn World.PACKAGE {id, ...} => Int.toString id)) pl

    fun ser b act = Int.toString b ^ " " ^ seract act

    val init = "Player"

    datatype move = MOVE of {robot : int,
                             pos : int * int,
                             picked : World.package list,
                             dropped : World.package list}

    type server = LSocket.socket

    val stoi = Option.valOf o Int.fromString

(*     fun hstoi s = Option.valOf (Int.fromString (String.extract (s, 1, NONE))) *)

    fun spotconv #"~" = SOME World.Water
      | spotconv #"." = SOME (World.Dot false)
      | spotconv #"*" = SOME (World.Dot false)
      | spotconv #"#" = SOME World.Wall
      | spotconv #"@" = SOME (World.Dot true)
      | spotconv c = NONE

    (* bow down before my glory --kaustuv *)
    (* your code uses so much memory that I am falling asleep waiting for it to terminate *)
    val makespots_slowly = 
      Array.fromList o List.concat o List.map (List.map ((fn space =>
                                                          World.SPOT {space = space, robots = []}) o
                                                         Option.valOf) o
                                               List.filter Option.isSome o
                                               List.map spotconv o String.explode)
    fun makespots w h sl =
	let val ss = Vector.fromList sl
	in Array.tabulate (w * h, fn i =>
			   let val col = i mod w
			       val row = i div w
			   in World.SPOT{space = valOf (spotconv (CharVector.sub (Vector.sub(ss, row), col))), 
					 robots = nil}
			   end)
	end


    fun updpacks loc packages nil = packages 
      | updpacks loc packages (id :: dsty :: dstx :: wt :: more) =
          updpacks loc (World.Map.insert (packages, id,
                                          World.PACKAGE {dest = World.Known (dstx, dsty),
                                                         weight = World.Known wt,
                                                         loc = World.Known (World.Ground loc),
                                                         id = id})) more

    fun locof robs id =
      (case World.Map.find (robs, id) of
         SOME (World.ROBOT {loc, ...}) => loc
       | NONE => raise Domain)

    fun movestart srv id dmap (World.BOARD {layout as (ll,ww,hh), cap, robots, packages}) =
      let
        val packs = (List.map stoi o String.tokens Char.isSpace) (recv srv)
        val loc = locof robots id
	val World.SPOT{space=sp, robots=rb} = World.sub layout loc

	fun clear_spots () = Array.modify (fn World.SPOT{space, ...} => 
					   World.SPOT{space=space, robots=nil}) ll
	fun add_rob loc rid = World.update layout loc (case World.sub layout loc of
							   World.SPOT{space, robots} => 
							       World.SPOT{space=space,
									  robots=rid::robots})

	fun filter nil = nil
          | filter (id :: dsty :: dstx :: wt :: more) =
	      if (FloodFill.readloc dmap layout (dstx,dsty) <> FloodFill.BIG) andalso wt <= (#1 cap)
	      then id::dsty::dstx::wt::(filter more)
	      else filter more
        val packs = filter packs
      in
	(* if no packages here, then make not a base *)
	if null packs then
	    (case sp of
		 World.Dot true => World.update layout loc (World.SPOT{space=World.Dot false, robots=rb})
	       | _ => ())
	else ();

	(* clear spots *)
	clear_spots ();

	(* fill in robots where they belong in spots *)
	World.Map.appi (fn (ri, World.ROBOT{loc,...}) => add_rob loc ri) robots;
	
        World.BOARD {layout = layout,
                     cap = cap,
                     robots = robots,
                     packages = updpacks loc packages packs}
      end

    fun dirmove (World.BOARD {layout, cap, robots, packages}) id dir =
      let
        val id = stoi id
        val World.ROBOT {loc = (row, col), money, packages = robotpacks, scorelb} = 
          Option.valOf (World.Map.find (robots, id))
      in
        World.BOARD {layout = layout,
                     cap = cap,
                     robots = World.Map.insert (robots, id,
                                                World.ROBOT {loc = (case dir of
                                                                      N => (row + 1, col)
                                                                    | S => (row - 1, col)
                                                                    | E => (row, col + 1)
                                                                    | W => (row, col - 1)),
                                                             money = money,
                                                             packages = robotpacks,
                                                             scorelb = scorelb}),
                     packages = packages}
      end
    
    fun pickdrops' id (packs, robots) pickp (pack :: more) =
      let
        val pack = stoi pack
        val robloc as (robr, robc) = locof robots id
        val loc = if pickp then World.Robot id else World.Ground robloc
        val packs = (case World.Map.find (packs, pack) of
                       SOME (World.PACKAGE {dest as World.Known (destr, destc), weight, id, ...}) => 
                         if pickp orelse not (destr = robr andalso destc = robc) then
                           World.Map.insert (packs, pack, World.PACKAGE {dest = dest,
                                                                         weight = weight,
                                                                         loc = World.Known loc,
                                                                         id = pack})
                         else #1 (World.Map.remove (packs, pack))
                     | SOME (World.PACKAGE {dest as World.Unknown, weight, id, ...}) => 
                           World.Map.insert (packs, pack,
                                             World.PACKAGE {dest = dest,
                                                            weight = weight,
                                                            loc = if pickp then World.Known loc
                                                                  else World.Unknown,
                                                            id = pack})
                     | NONE => 
                         if pickp then 
                           World.Map.insert (packs, pack, World.PACKAGE {dest = World.Unknown,
                                                                         weight = World.Unknown,
                                                                         loc = World.Known loc,
                                                                         id = pack})
                         else packs)
      in
        pickdrops id (packs,
                      World.Map.mapi (fn (rid, rob as World.ROBOT {loc, money, packages, scorelb}) =>
                                      if rid = id then 
                                        World.ROBOT {loc = loc, money = money,
                                                     packages =
                                                       if pickp then pack :: packages
                                                       else List.filter (fn p => p <> pack) packages,
                                                      scorelb = scorelb}
                                      else rob) robots) more
      (* XXX deal with scorelb later *)
      end
    and pickdrops id (packs, robots) ("P" :: more) = pickdrops' id (packs, robots) true more
      | pickdrops id (packs, robots) ("D" :: more) = pickdrops' id (packs, robots) false more
      | pickdrops id (packs, robots) more = (packs, robots, more)

    fun packacts (brd as World.BOARD {layout, cap, robots, packages}) id more =
      let
        val id = stoi id
        val (packs, robs, more) = pickdrops id (packages, robots) more
      in
        (World.BOARD {layout = layout,
                      cap = cap,
                      packages = packs,
                      robots = robs}, more)
      end
    
    fun makeappear (World.BOARD {layout, cap = (carry, mon), robots, packages}) id (x, y) =
      let
        val (x, y) = (stoi x, stoi y)
        val id = stoi id
      in
        World.BOARD {layout = layout, cap = (carry, mon),
                     robots = World.Map.insert (robots, id, World.ROBOT {loc = (x, y),
                                                                         money = mon,
                                                                         packages = [],
                                                                         scorelb = 0}), (* XXX *)
                     packages = packages}
      end
                                                                         
        
    fun makemoves brd (id, "N" :: more) = makemoves (dirmove brd id N) (id, more)
      | makemoves brd (id, "S" :: more) = makemoves (dirmove brd id S) (id, more)
      | makemoves brd (id, "E" :: more) = makemoves (dirmove brd id E) (id, more)
      | makemoves brd (id, "W" :: more) = makemoves (dirmove brd id W) (id, more)
      | makemoves brd (id, picks as "P" :: _) =
        let val (brd, more) = packacts brd id picks in makemoves brd (id, more) end
      | makemoves brd (id, drops as "D" :: _) =
        let val (brd, more) = packacts brd id drops in makemoves brd (id, more) end
      | makemoves brd (id, "X" :: x :: "Y" :: y :: more) =
             makemoves (makeappear brd id (y, x)) (id, more)
      | makemoves brd (id, []) = brd
      | makemoves brd (id, more) = (pr ["Cannot match: ", StringUtil.delimit " " more, "\n"];
                                    raise Domain)
    fun uncons nil = raise Domain
      | uncons (x :: xs) = (x, xs)

(*     fun minus X [] = X *)
(*       | minus X (h :: t) = minus (List.filter (fn x => x <> h) X) t *)

(*     fun keys M = List.map (fn (x, _) => x) (World.Map.listItemsi M) *)

    fun keeponly (World.BOARD {layout, cap, robots, packages}) ids =
      World.BOARD {layout = layout,
                   cap = cap,
                   packages = packages,
                   robots = World.Map.filteri (fn (rid, rob) => List.exists (fn j => j = rid) ids) robots}

    fun move (srv : server) me (bid : bid, act : action) (World.BOARD{layout,cap,robots,packages})  =
      let
        val sent = ser bid act
        val _ = if !debug then pr ["sending move: ", sent, "\n"] else ()
        val _ = LSocket.send srv (ser bid act)
	    
	val newrobots = World.Map.mapi (fn (id, bot as World.ROBOT{loc,money,packages,scorelb}) =>
					    if id = me then World.ROBOT{loc=loc, money = money - bid,
									packages=packages, scorelb=scorelb}
					    else bot) robots

	val brd = World.BOARD{layout=layout, cap=cap, robots=newrobots, packages=packages}

        val rcv = recv srv
        val upds = String.tokens (fn x => Char.compare (x, #"#") = EQUAL) rcv
        val _ = if !debug then pr ["starting makemoves at: ", Time.toString (Time.now ()), "\n"] else ()
        val newbrd = List.foldr (fn (m, b) => makemoves b (uncons (String.tokens Char.isSpace m))) brd upds
        val newbrd = keeponly newbrd (List.map (stoi o List.hd o String.tokens Char.isSpace) upds)
        val _ = if !debug then pr ["makemoves done at: ", Time.toString (Time.now ()), "\n"] else ()
        val _ = if !showboard then 
	    (pr ["Board = \n"];
	     Util.time "pr_board" pr_board newbrd;
	     pr ["\n"]) else ()
      in
	  if !pause then
	      (TextIO.print "? ";
	       TextIO.inputLine TextIO.stdIn;
	       ())
	  else ();
(*        Posix.Process.sleep (Time.fromMilliseconds 1000); *)
        newbrd
      end

    fun contact (hn, port) = LSocket.connect hn port

      
    fun init srv =
      let
        val _ = LSocket.send srv "Player"
        val wid :: hei :: _ = String.tokens Char.isSpace (recv srv)
        val (wid, hei) = (stoi wid + 2, stoi hei + 2)
        val spots =
          (String.implode (List.tabulate (wid, fn _ => #"#"))
           :: List.tabulate (hei - 2, fn _ => "#" ^ recv srv ^ "#"))
          @ [String.implode (List.tabulate (wid, fn _ => #"#"))]
        val id :: cap :: mon :: _ = String.tokens Char.isSpace (recv srv)
        val (id, cap, mon) = (stoi id, stoi cap, stoi mon)
        val inits = String.tokens (fn x => Char.compare (x, #"#") = EQUAL) (recv srv)

        val robs = World.Map.insert (World.Map.empty,
                                     id, 
                                     World.ROBOT {loc = (~1, ~1),
                                                  money = mon,
                                                  packages = [],
                                                  scorelb = 0})
        val packs = World.Map.empty
        val _ = if !debug then pr ["attempting to parse the board", "\n"] else ()
        val spots = (Util.time "makespots" (makespots wid hei)) spots
	val _ = if !debug then pr ["done..\n"] else ()
        val board = World.BOARD {layout = (spots, wid, hei),
                                 cap = (cap, mon),
                                 robots = robs,
                                 packages = packs}

        val board = List.foldr (fn (m, b) => makemoves b (uncons (String.tokens Char.isSpace m))) board inits


(*         val brd = makemoves board (uncons inits) *)
        val _ = if !showboard then pr_board board else ()
      in
        {board = board, me = id}
      end

    fun disconnect s = LSocket.hangup s
  end
  
