structure Knapsack =
struct
    open World

    val debug = Params.flag false (SOME ("-debug",
                                         "Turn on debugging messages.")) "debug"

    type rankedpkg = real * package
    type cluster = real * (int * int * package list) (* Score, weight, distance, packages *)

    val overcarry_limit = 8

    (* Sort rankedpkg or cluster according to descending score *)
    fun sort l = rev (ListUtil.sort (ListUtil.byfirst Real.compare) l)

    fun score (p as PACKAGE{dest as Known(r,c), weight as Known w,...}) layout dmap =
	Real.fromInt w / Real.fromInt(1 + FloodFill.readloc dmap layout (r,c))

    fun rawscore (w,d) layout dmap = (Real.fromInt w) / (Real.fromInt (d+1))

    fun pkgweight (PACKAGE{weight as Known w,...}) = w
    fun rankedpkgweight (_,PACKAGE{weight as Known w,...}) = w

    fun clusterweight (_,(w,_,_)) = w

    fun getscore nil = 0.0
      | getscore ((s,_)::t) = s+(getscore t)

    fun getweight nil = 0
      | getweight ((_,PACKAGE{weight as Known w,...})::t) = w + (getweight t)

    (* Helper to partition a list along the boundary of what can be carried.
       Returns a tuple of (must be carried, might be carried)
       partition_weight : ('a -> int) -> int -> 'a list -> ('a list * 'a list) *)
    fun partition_weight f weight l =
      let
        fun take' (l,len) = List.take (l,len) handle _ => l
        fun drop' (l,len) = List.drop (l,len) handle _ => nil

        fun partition' must nil _ = (must,nil)
          | partition' must (might as (h::t)) w = if w<0
              then (drop' (must,overcarry_limit),
		    (take' (must,overcarry_limit)) @ (take' (might,overcarry_limit)))
	      else partition' (h::must) t (w-(f h))
      in
        partition' nil l weight
      end

    (* build_rankedpkg : spot Array.array * int * int) ->
       FloodFill.distmap -> package -> rankedpkg *)
    fun build_rankedpkg layout dmap p = (score p layout dmap, p)

    (* build_list : (spot Array.array * int * int) -> FloodFill.distmap ->
       package list -> rankedpkg list (sorted) *)
    fun build_list layout dmap l = sort (map (build_rankedpkg layout dmap) l)

    (* build_cluster : board -> layout -> distmap -> distmap -> cluster -> rankedpkg list ->
       cluster * rankedpkg list
       First distmap is robot distances, second is for use in the function *)
    fun build_cluster board layout botmap dmap (cls as (_,(_,_,pkg::t))) l =
      let
        val threshhold = 3 (*score pkg layout botmap*)
	val dist = FloodFill.readloc dmap layout

        (* build' : cluster -> rankedpkg list -> cluster * rankedpkg list *)
        fun build' (cls : cluster) nil = (cls,nil)
	  | build' (cls as (s,(w,d,l as (pkg as PACKAGE{dest as Known dest,...})::_))) (h::t) =
	  let
	    val _ = FloodFill.updatemap dmap board dest
            (* Find the closest package in foldl (below) *)
            fun test_dst (p' as (_,PACKAGE{dest as Known dest,...}),pos) =
		if dist dest < dist pos then dest else pos
	    val (_,PACKAGE{dest = Known firstpos,...}) = h
	    val bestpos = foldl test_dst firstpos t
	  in
	    if dist bestpos <= threshhold then
              let
                val (inc, outc) = List.partition (fn (_,PACKAGE{dest=Known d,...}) => bestpos=d) (h::t)
                val inc' = #2 (ListPair.unzip inc)
              in build' (rawscore (w+(getweight inc),d) layout botmap,(w+(getweight inc),d,l@inc')) outc
              end
            else (cls, h::t)
          end

        val (cls', pl) = build' cls l
        val pl' = build_list layout botmap (#2 (ListPair.unzip pl))
      in
	(cls',pl')
      end

    exception Unfinished
    (* pickup : board -> layout -> FloodFill.distmap -> package list -> int -> package list *)
    fun pickup board layout botmap pkgs limit =
      let
	  val pkgs' = build_list layout botmap pkgs
	  val dmap = FloodFill.makemap board

	  val dist = FloodFill.readloc dmap layout
 
	  (* get' : cluster list -> rankedpkg list -> int -> int -> cluster list
             ints are weight remaining and post-weight countdown *)
	  fun get' cl nil _ _ = cl
	    | get' cl _ _ 0 = cl
            | get' cl (pl as (_,p as PACKAGE{dest as Known d,...})::t) weight countdown =
            let
	      val cd = if weight<=0 then countdown-1 else countdown
              val (inc,outc) = List.partition (fn (_,PACKAGE{dest as Known d',...}) => d = d') pl
	      val (s,w) = (getscore inc,getweight inc)
	      val (cls as (s,(w',d,l)),outc') =
                  build_cluster board layout botmap dmap (s,(w, FloodFill.readloc botmap layout d,
                    #2 (ListPair.unzip inc))) outc
	      val w' = Int.min(w',limit)
	      val cls = (rawscore (w',d) layout botmap,(w',d,l))
            in
              get' (cls::cl) outc' (weight-w') cd
            end

          (* Brute force decision function, given a weight function f *)
          fun decide f l limit =
            let
              fun best nil _ = (0.0,nil,nil)
                | best ((h as (s,_))::t) lim =
                  let
                    val (noval, nolist, nolist') = best t lim
                  in
                    if f h <= lim then
                      let
			val (yesval, yeslist, yeslist') = best t (lim-(f h))
                      in
			if yesval+s >= noval then (yesval+s,h::yeslist,yeslist')
                        else (noval,nolist,h::nolist')
                      end
                    else (noval,nolist,h::nolist')
                  end
                val (_,lst,lst') = best l limit
	      in
		  (lst,lst')
	      end

	  fun pkgweights nil = 0
	    | pkgweights (h::t) = (pkgweight h) + (pkgweights t)

	  fun greedy nil _ = (if !debug then print "greedy called on nil\n" else (); nil)
            | greedy (h::t) lim = 
              let in
                if (pkgweight h) <= lim then
                  (if !debug then let val (PACKAGE{id,...}) = h in print ("Capacity " ^ (Int.toString lim) ^
		      ", adding package " ^ (Int.toString id) ^ " for weight addition of " ^
		      (Int.toString (pkgweight h)) ^ "\n") end else ();
                  h :: (greedy t (lim-(pkgweight h))))
	        else greedy t lim
              end

	  val candidates = sort (get' nil pkgs' limit overcarry_limit)

	  fun sortcls cl nil = cl
	    | sortcls nil (h::t) = sortcls [h] t
	    | sortcls (cl as ((_,(_,_,PACKAGE{dest = Known dest,...}::_))::_)) (h::t) =
	      let
		val _ = FloodFill.updatemap dmap board dest
		val (_,(_,_,PACKAGE{dest = Known destin,...}::_)) = h
		val (c,_,l) = foldl (fn (c as (_,(_,_,((PACKAGE{dest = Known dest',...})::_))),(c',d,l)) =>
		     if (dist dest') < (dist d) then (c,dest',c'::l) else (c',d,c::l)) (h,destin,nil) t
	      in sortcls (c::cl) l
	      end
	  val packages = foldl (fn ((_,(_,_,l)),l') => l'@l) nil (rev (sortcls nil candidates))


	  val (must,might) = partition_weight pkgweight limit packages

	  val (decision,_) = decide rankedpkgweight (map (build_rankedpkg layout dmap) might) (limit-(pkgweights must))
      in
	  must@(#2 (ListPair.unzip decision))
      end
end

(*structure Knapsack =
  struct
    val overcarry_limit = 8

    fun sort l = rev (ListUtil.sort (ListUtil.byfirst Real.compare) l)

    (* Package information must be known *)
    fun makelist pl layout dm =
        map (fn (p as World.PACKAGE{dest as World.Known(r,c),
          weight as World.Known w,...}) =>
          (Real.fromInt w / Real.fromInt(FloodFill.readloc dm layout (r,c)), p))
          pl

    (* pickup World.package list -> int -> (spot Array.array * int * int) ->
         FloodFill.distmap -> World.package list *)
    fun pickup pkgs limit layout distmap =
      let
        fun take' (l,len) = List.take (l,len) handle _ => l
        fun drop' (l,len) = List.drop (l,len) handle _ => nil

        val pkgs = sort (makelist pkgs layout distmap)

	(* (real*package) list * (real*package) list * int ->
           (real*package) list * (real*package) list *)
        (* candidates takes in (list of possible packages, nil, weight limit) and
           returns a tuple of two lists. The first list is packages we definitely
           want to pick up (greedily chosen from the beginning of the sorted list),
           and the second list is overcarry_limit*2 packages that we might want to
           pick up. These should then be run through get_best to choose the list of
           candidates that will actually be used. *)
	fun candidates (carrying : ((real*World.package) list)) nil lim = (carrying,nil)
          | candidates (carrying : ((real*World.package) list)) ((score : real,
              p as World.PACKAGE{weight as (World.Known w),...})::l) lim =
              if lim < 0 then (drop' (carrying,overcarry_limit),
                               (take' (carrying,overcarry_limit)) @
                               (take' ((score,p)::l,overcarry_limit)))
              else candidates ((score,p)::carrying) l (lim-w)

        val totalweight = foldl
          (fn (World.PACKAGE{weight as World.Known w,...},t) => t+w) 0

        val (get, try) = candidates nil pkgs limit
        val tryweight = limit - (totalweight (#2 (ListPair.unzip get)))

	fun get_best nil _ = (0.0,nil)
	  | get_best ((v,p as World.PACKAGE{weight as World.Known w,...})::l)
	      lim =
            let
              val (noval, nolist) = get_best l lim
            in
              if w <= lim then
                let
                  val (yesval, yeslist) = get_best l (lim-w)
                in
		  if yesval >= noval then (yesval+v,p::yeslist)
		  else (noval, nolist)
                end
              else (noval, nolist)
            end
      in
        (#2 (ListPair.unzip get)) @ (#2 (get_best try limit))
      end
  end*)